/*
*******************************************************************************************************************
*								SHT10 sensor demonstration program, test on dummy data
* Description:  This program demonstrates basic functions to use SHT10 sensor.
*				Use Docklight to check results. Baud rate is 9600bps. Please finger close to
*				temperature and RH sensor and note the values changed reflected in DockLight program.
* File:			SHTxx_Sample_Code.c
* Software:		Keil C (C51) compiler, Eval version 7.09 under uVision2 IDE v2.40
* Programmer: 	John Leung, TechToys Co. (www.TechToys.com.hk)
* Hardware:		AT89S52-gLCD-STK1, microcontroller = AT89C51RB2, Vdd=3.3V
* Date:			2nd Feb 2007
* Version:		1.0
*******************************************************************************************************************
*/

#include <REG51RB2.H>	//Microcontroller specific library, e.g. port definitions
#include <intrins.h> 	//Keil library (is used for _nop()_ operation)  
#include <stdio.h>   	//Keil library

//----------------------------------------------------------------------------------
// modul-var
//----------------------------------------------------------------------------------
enum {TEMP,HUMI};

sbit	DATA=P3^5;
sbit	SCK =P3^4;

#define noACK 0
#define ACK   1
                            //adr  command  r/w
#define STATUS_REG_W 0x06   //000   0011    0
#define STATUS_REG_R 0x07   //000   0011    1
#define MEASURE_TEMP 0x03   //000   0001    1
#define MEASURE_HUMI 0x05   //000   0010    1
#define RESET        0x1e   //000   1111    0

//----------------------------------------------------------------------------------
char s_write_byte(unsigned char value)
//----------------------------------------------------------------------------------
// writes a byte on the Sensibus and checks the acknowledge 
{ 
  unsigned char i,error=0;  
  for (i=0x80;i>0;i/=2)             //shift bit for masking
  { if (i & value) DATA=1;          //masking value with i , write to SENSI-BUS
    else DATA=0;                        
    SCK=1;                          //clk for SENSI-BUS
    _nop_();_nop_();_nop_();        //pulswith approx. 5 us  	
    SCK=0;
  }
  DATA=1;                           //release DATA-line
  SCK=1;                            //clk #9 for ack 
  error=DATA;                       //check ack (DATA will be pulled down by SHT11)
  SCK=0;        
  return error;                     //error=1 in case of no acknowledge
}

//----------------------------------------------------------------------------------
char s_read_byte(unsigned char ack)
//----------------------------------------------------------------------------------
// reads a byte form the Sensibus and gives an acknowledge in case of "ack=1" 
{ 
  unsigned char i,val=0;
  DATA=1;                           //release DATA-line
  for (i=0x80;i>0;i/=2)             //shift bit for masking
  { SCK=1;                          //clk for SENSI-BUS
    if (DATA) val=(val | i);        //read bit  
    SCK=0;  					 
  }
  DATA=!ack;                        //in case of "ack==1" pull down DATA-Line
  SCK=1;                            //clk #9 for ack
  _nop_();_nop_();_nop_();          //pulswith approx. 5 us 
  SCK=0;						    
  DATA=1;                           //release DATA-line
  return val;
}

//----------------------------------------------------------------------------------
void s_transstart(void)
//----------------------------------------------------------------------------------
// generates a transmission start 
//       _____         ________
// DATA:      |_______|
//           ___     ___
// SCK : ___|   |___|   |______
{  
   DATA=1; SCK=0;                   //Initial state
   _nop_();
   SCK=1;
   _nop_();
   DATA=0;
   _nop_();
   SCK=0;  
   _nop_();_nop_();_nop_();
   SCK=1;
   _nop_();
   DATA=1;		   
   _nop_();
   SCK=0;		   
}

//----------------------------------------------------------------------------------
void s_connectionreset(void)
//----------------------------------------------------------------------------------
// communication reset: DATA-line=1 and at least 9 SCK cycles followed by transstart
//       _____________________________________________________         ________
// DATA:                                                      |_______|
//          _    _    _    _    _    _    _    _    _        ___     ___
// SCK : __| |__| |__| |__| |__| |__| |__| |__| |__| |______|   |___|   |______
{  
  unsigned char i; 
  DATA=1; SCK=0;                    //Initial state
  for(i=0;i<9;i++)                  //9 SCK cycles
  { SCK=1;
    SCK=0;
  }
  s_transstart();                   //transmission start
}
		
				   
//----------------------------------------------------------------------------------
char s_measure(unsigned char *p_value, unsigned char *p_checksum, unsigned char mode)
//----------------------------------------------------------------------------------
// makes a measurement (humidity/temperature) with checksum
{ 
  unsigned error=0;
  unsigned int i;

  s_transstart();                   //transmission start
  switch(mode){                     //send command to sensor
    case TEMP	: error+=s_write_byte(MEASURE_TEMP); break;
    case HUMI	: error+=s_write_byte(MEASURE_HUMI); break;
    default     : break;	 
  }
  for (i=0;i<65535;i++) if(DATA==0) break; //wait until sensor has finished the measurement
  if(DATA) error+=1;                // or timeout (~2 sec.) is reached
  *(p_value)  =s_read_byte(ACK);    //read the first byte (MSB)
  *(p_value+1)=s_read_byte(ACK);    //read the second byte (LSB)
  *p_checksum =s_read_byte(noACK);  //read checksum
  return error;
}

//----------------------------------------------------------------------------------
void init_uart()
//----------------------------------------------------------------------------------
//9600 bps @ 11.059 MHz 
{SCON  = 0x52;    
 TMOD  = 0x20;    
 TCON  = 0x69;	  
 TH1   = 0xfd;    
}

//See an application note Non-linearity Compensation from www.Sensirion.com for details
//----------------------------------------------------------------------------------------
void calc_sht10(unsigned int *p_humidity ,int *p_temperature)
//----------------------------------------------------------------------------------------
// calculates temperature [Deg Cel] and humidity [%RH] 
// input :  humi [Ticks] (12 bit) 
//          temp [Ticks] (14 bit)
// output:  humi [10x %RH] for 1 decimal place
//          temp [100x Deg Cel] for 2 decimal places
{ 
	unsigned int rh = *p_humidity;
	int t 	= *p_temperature;
	int t_C;
	unsigned int rh_lin;

	//t_C=t*0.01-40
	//multiplied by 100, t_C=t-4000
	t_C = t - 4000;

	//58<=rh<=1720, 	(1430*rh-5120*16)/4096, 	in x10
	//1721<=rh<=3273, 	(1110*rh+28930*16)/4096, 	in x10
	//From : Application Note Non-Linearity compensation, Sensirion
	if(rh<=1720)
	{
		rh_lin=(1430UL*(long)rh)>>12;
		(rh_lin>=20)?(rh_lin = rh_lin-20):(rh_lin=0);
	} 
	else
	{
		//rh_lin = (1110*rh + 12320 + 4096*110)/4096, in x10
		rh_lin=((1110UL*(long)rh+12320UL)>>12)+110;
		if(rh_lin>1000) rh_lin=1000;
	}

	//rh_true = (t_C-25)*(rh*0.00008+0.01)+rh_lin
	//~0.12 %RH /deg C, this factor has been ignored at this moment
	*p_temperature = t_C;
	*p_humidity = rh_lin;
}

//----------------------------------------------------------------------------------
void main()
//----------------------------------------------------------------------------------
// Sample program that shows how to use SHT10 functions
// 1. connection reset 
// 2. measure humidity [ticks](12 bit) and temperature [ticks](14 bit)
// 3. calculate humidity x10 [%RH] and temperature x100 [Deg Cel]
// 4. print temperature, humidity in 1 decimal place

{ unsigned int humi_val;
  int temp_val;
  unsigned char error,checksum;
  unsigned int i;

  init_uart();

  s_connectionreset();
  while(1)
  {
	error=0;
    error+=s_measure((unsigned char*) &humi_val,&checksum,HUMI);  //measure humidity
    error+=s_measure((unsigned char*) &temp_val,&checksum,TEMP);  //measure temperature
    if(error!=0) s_connectionreset();                 	//in case of an error: connection reset
    else
    { 
      calc_sht10(&humi_val,&temp_val);            		//calculate humidity, temperature
      printf("temp: %d.%u  humidity: %d.%u\n", temp_val/100, (temp_val)%100/10, humi_val/10, humi_val%10);
    }

    //----------wait approx. 0.8s to avoid heating up SHTxx------------------------------      
    for (i=0;i<40000;i++);     //(be sure that the compiler doesn't eliminate this line!)
    //-----------------------------------------------------------------------------------                       
  }

} 
